/*
 * JBoss, Home of Professional Open Source
 * Copyright 2008, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags.
 * See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions
 * of the GNU Lesser General Public License, v. 2.1.
 * This program is distributed in the hope that it will be useful, but WITHOUT A
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License,
 * v.2.1 along with this distribution; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * (C) 2008
 * @author JBoss Inc.
 */
package org.jboss.jbossts.rts.service;

import org.jboss.jbossts.rts.resource.TransactionList;
import org.jboss.jbossts.rts.resource.Transaction;
import org.jboss.jbossts.rts.provider.ResourceNotFoundException;
import org.jboss.jbossts.rts.provider.TransactionStatusException;
import org.jboss.jbossts.rts.util.TxUtil;

import javax.ws.rs.*;
import javax.ws.rs.core.Response;
import org.jboss.resteasy.util.HttpResponseCodes;

import java.util.HashMap;
import java.util.ArrayList;
import java.util.List;
import org.apache.log4j.Logger;

import com.arjuna.ats.arjuna.coordinator.ActionStatus;
import com.arjuna.ats.arjuna.AtomicAction;
import com.arjuna.ats.arjuna.common.Uid;

@Path("/")
public class Coordinator
{
    protected final static Logger log = Logger.getLogger(Coordinator.class);

    private HashMap<String, Transaction> transactions = new HashMap<String, Transaction>();
    // each participant may only be enlisted in one transaction
    private HashMap<String, String> participants = new HashMap<String, String>();

    /**
     * Performing a GET on /transaction-coordinator returns a list of all transactions know to
     * the coordinator (active and recovery).
     * @return list of all transactions know to the coordinator (active and recovery).
     */
    @GET
    @Path("transaction-coordinator")
    @Produces("application/json")
    public TransactionList getAllTransactions()
    {
        log.trace("coordinator: list: transaction-coordinator");
        return getTransactions(null);
    }

    /**
     * Lookup the status of a single transaction.
     * @param id the transaction id
     * @return a list of transactions containing the status of the requested item
     *  (which will be empty if the tx does not exist)
     */
    @GET
    @Path("transaction-coordinator/{id}")
    @Produces("application/json")
    public TransactionList getTransactionStatus(@PathParam("id") String id)
    {
        log.trace("coordinator: status: transaction-coordinator/" + id);
        final Uid uid = new Uid(id);
        TransactionFilter statusFilter = new TransactionFilter() {
            public boolean accept(Transaction t)
            {
                return uid.equals(t.get_uid());
            }
        };

        return getTransactions(statusFilter);
    }

    @GET
    @Path("transaction-coordinator.html")
    @Produces("text/html")
    public Response getAllTransactionsText() throws Exception
    {
        log.trace("coordinator: html list");
        TransactionList listing = getTransactions(null);

        return Response.ok(TxUtil.stringFormat(listing)).build();
    }

    /**
     * Perform a GET on /transaction-coordinator/recovery
     * @return a list of transactions that are in recovery.
     */
    @GET
    @Path("transaction-coordinator/recovery")
    @Produces("application/json")
    public TransactionList getRecoveringTransactions()
    {
        log.trace("coordinator: transaction-coordinator/recovery");
        return getTransactions(recoveringFilter);
    }

    /**
     * Perform a GET on /transaction-coordinator/active
     * @return a list of inflight transaction ids
     */
    @GET
    @Path("transaction-coordinator/active")
    @Produces("application/json")
    public TransactionList getActiveTransactions()
    {
        log.trace("coordinator: transaction-coordinator/active");
        return getTransactions(activeTxFilter);
    }
    /**
     * Performing a DELETE on any of the transaction-coordinator URLs will return a 401.
     * @return 401
     */
    @DELETE
    @Path("transaction-coordinator/{id}")
//    @RolesAllowed(value = "NONE")
    public Response deleteTransaction(@PathParam("id") String id)
    {
        log.trace("coordinator: delete: transaction-coordinator/" + id);
        return TxUtil.response(HttpResponseCodes.SC_UNAUTHORIZED);
    }

    /**
     * Start a new transaction (which will be deleted when the tx timeout value is reached).
     * Calls to a timed out tx will return 404 (in which case the caller may assume rollback)
     * @param clientId the unique identity of the client (can be a URL too)
     * @param timeout period in milliseconds
     * @return a <URL> of the form /transaction-coordinator/<TxId>
     */
    @POST
    @Path("transaction-coordinator/begin/")
    public Response beginTransaction(@QueryParam("clientId")String clientId, @QueryParam("timeout") @DefaultValue("-1")int timeout)
//    @Path("transaction-coordinator/begin/{clientId}")
//    public Response beginTransaction(@PathParam("clientId")String clientId, @QueryParam("timeout") @DefaultValue("0")int timeout)
    {
        log.trace("coordinator: begin: transaction-coordinator/begin/" + clientId + "?timeout=" + timeout);
        Transaction tx = new Transaction(clientId);
        String uid = tx.get_uid().fileStringForm();

        transactions.put(uid, tx);
        int status = tx.begin(timeout);

        try
        {
            if (status == ActionStatus.RUNNING)
                return TxUtil.postResponse(TxUtil.txURI + '/' + uid);

            throw new TransactionStatusException("Transaction failed to start: " + status);
        }
        finally
        {
            AtomicAction.suspend();
        }
    }

    /**
     * trigger a commit of the transaction. Upon termination, the URL is implicitly
     * deleted. If it no longer exists, then 404 will be returned. The invoker
     * cannot know for sure whether the transaction committed or rolled back.
     * This is the same problem with JTS: to be sure you should register a 2PC participant.
     * @param txId transaction id
     * @return http response code
     */
    @PUT
    @Path("transaction-coordinator/{TxId}/commit")
    public Response commitTransaction(@PathParam("TxId")String txId, @QueryParam("fault") @DefaultValue("")String fault)
    {
        log.debug("coordinator: commit: transaction-coordinator/" + txId + "/commit");
        Transaction tx = getTransaction(txId);

        tx.setFault(fault);
        AtomicAction.resume(tx);
        int res = tx.commit(true);
        log.trace("commit result: " + res);

        if (tx.isTransactionInMidFlight())
        {
            AtomicAction.suspend();
            throw new TransactionStatusException("Transaction failed to commit: ");
        }

        if (!tx.hasHeuristic())
            transactions.remove(txId);

        return TxUtil.response(HttpResponseCodes.SC_OK);
    }

    @PUT
    @Path("transaction-coordinator/{TxId}/rollback")
    public Response rollbackTransaction(@PathParam("TxId")String txId, @QueryParam("fault") @DefaultValue("")String fault)
    {
        log.debug("coordinator: rollback: transaction-coordinator/" + txId + "/rollback");
        Transaction tx = getTransaction(txId);

        tx.setFault(fault);
        AtomicAction.resume(tx);
        int res = tx.abort();
        log.trace("rollback result: " + res);

        if (tx.isTransactionInMidFlight())
        {
            AtomicAction.suspend();
            throw new TransactionStatusException("Transaction failed to rollback: ");
        }

        if (!tx.hasHeuristic())
            transactions.remove(txId);

        return TxUtil.response(HttpResponseCodes.SC_OK);
    }

    /**
     * Register a participant in a tx
     * @param txId id of transaction
     * @param url participant id (unique within the scope of txId)
     * @return unique resource ref for the participant
     */
    @PUT
    @Path("transaction-coordinator/{TxId}")
    @Consumes("text/html")
    public Response enlistParticipant(@PathParam("TxId")String txId, @QueryParam("URL")String url, String content)
    {
        Transaction tx = getTransaction(txId);

        if (url == null)
		url = content.trim();

        String coordinatorId = tx.enlistParticipant(url);

        if (coordinatorId == null)
            return TxUtil.response(HttpResponseCodes.SC_NOT_ACCEPTABLE);

        log.debug("enlisting participant: content=" + content +
		" URL=" + url + " in tx " + txId + " Coordinator url: " + TxUtil.rcURI + '/' + coordinatorId);
        // TODO would be nice if there was something in the spec for telling the participant it
        // has been enlisted
        participants.put(coordinatorId, url);

        return Response.ok(TxUtil.rcURI + '/' + coordinatorId).build();
    }
         
    /**
     * Get the participant url (registered during enlistParticipant) corresponding to a resource reference
     * if the coordinator crashes - the participant list will be empty but this is ok if commit hasn't been
     * called since the TM uses presumed abort sematics.
     * @param enlistmentId the resource reference
     * @return the participant url
     */
    @GET
    @Path("recovery-coordinator/{RecCoordId}")
    public Response lookupParticipant(@PathParam("RecCoordId")String enlistmentId)
    {
        log.trace("coordinator: lookup: transaction-coordinator/" + enlistmentId);

        String p = participants.get(enlistmentId);

        if (p == null)
            return TxUtil.response(HttpResponseCodes.SC_NOT_FOUND);

        return Response.ok(p).build();
    }

    /**
     * PUT /recovery-coordinator/<RecCoordId>/<new participant URL> -
     *   overwrite the old <participant URL> with <new participant URL>
     *   (as with JTS, this will also trigger off a recovery attempt on the associated transaction)
     * A participant may use this url to notifiy the coordinator that he has moved to a new location.
     * @param url
     */
    @PUT
    @Path("recovery-coordinator/{RecCoordId}")
    public Response replaceParticipant(@PathParam("RecCoordId")String enlistmentId, @QueryParam("URL")String url)
    {
        log.trace("coordinator: replace: recovery-coordinator/" + enlistmentId + "?URL=" + url);
        participants.put(enlistmentId, url);

        return TxUtil.response(HttpResponseCodes.SC_OK);
    }

    @POST
    @Path("recovery-coordinator/{RecCoordId}")
    public Response postParticipant(@PathParam("RecCoordId")String enlistmentId)
    {
        log.trace("coordinator: replace via Post: recovery-coordinator/" + enlistmentId);
        return Response.status(HttpResponseCodes.SC_UNAUTHORIZED).build();
    }

    @DELETE
    @Path("recovery-coordinator/{RecCoordId}")
    public Response deleteParticipant(@PathParam("RecCoordId")String enlistmentId)
    {
        log.trace("coordinator: replace vi Delete: recovery-coordinator/" + enlistmentId);
        return TxUtil.response(HttpResponseCodes.SC_UNAUTHORIZED);
    }

    private Transaction getTransaction(String txId)
    {
        Transaction tx = transactions.get(txId);

        if (tx == null)
            throw new ResourceNotFoundException("Transaction id not found");

        return tx;
    }

    private TransactionList getTransactions(TransactionFilter filter)
    {
        List<Transaction> list = new ArrayList<Transaction>();

        if (filter == null)
        {
            list.addAll(transactions.values());
        }
        else
        {
            for (Transaction t : transactions.values())
                if (filter.accept(t))
                    list.add(t);
        }

        return new TransactionList(list);
    }

    TransactionFilter activeTxFilter = new TransactionFilter() {
        public boolean accept(Transaction t)
        {
            return t.isTransactionInMidFlight();
        }
    };

    TransactionFilter recoveringFilter = new TransactionFilter() {
        public boolean accept(Transaction t)
        {
            return !t.isTransactionInMidFlight();
        }
    };

    interface TransactionFilter
    {
        boolean accept(Transaction t);
    }
}

