/*
 * JBoss, Home of Professional Open Source
 * Copyright 2008, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags.
 * See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions
 * of the GNU Lesser General Public License, v. 2.1.
 * This program is distributed in the hope that it will be useful, but WITHOUT A
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License,
 * v.2.1 along with this distribution; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * (C) 2008
 * @author JBoss Inc.
 */
package org.jboss.jbossts.rts.util;

import org.apache.commons.httpclient.methods.*;
import org.apache.commons.httpclient.*;
import org.apache.log4j.Logger;
import org.jboss.jbossts.rts.provider.HttpResponseException;
import org.jboss.resteasy.plugins.providers.jaxb.json.JettisonMappedContext;
import org.jboss.resteasy.util.HttpResponseCodes;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;

import javax.xml.bind.Marshaller;
import javax.xml.bind.JAXBException;
import javax.ws.rs.core.Response;

/**
 * Various utilites for sending HTTP messages
 */
public class TxUtil
{
    protected static final Logger log = Logger.getLogger(TxUtil.class);
    
    public static final String TX_URI_PREFIX = "transaction-coordinator";
    public static final String RC_URI_PREFIX = "recovery-coordinator";
    
    private static HttpClient client = new HttpClient(new MultiThreadedHttpConnectionManager());

    private static int serverPort = getIntProperty("PORT", "9096");
    private static String serverHost = "localhost";
    private static String serverContext = "tx/";
    private static String baseURI = "http://" + serverHost + ":" + serverPort + "/" + serverContext;

    public static String txURI = TxUtil.getBaseURI() + TX_URI_PREFIX;
    public static String rcURI = TxUtil.getBaseURI() + RC_URI_PREFIX;

    public static int getIntProperty(String name, String def)
    {
        String port = System.getProperty(name, def);

        try {
            return Integer.valueOf(port);
        } catch (NumberFormatException e) {
            try {
                return Integer.valueOf(def);
            } catch (NumberFormatException e2) {
                throw new IllegalArgumentException("invalid default value for server port: " + def, e2);
            }
        }
    }

    private static void updateEndpoint()
    {
        baseURI = "http://" + serverHost + ":" + serverPort + "/" + serverContext;
        txURI = baseURI + TX_URI_PREFIX;
        rcURI = baseURI + RC_URI_PREFIX;
        log.info("base URI: " + baseURI);
    }

    public static void setPort(int port)
    {
        serverPort = port;
        updateEndpoint();
    }

    public static void setHost(String host)
    {
        serverHost = host;
        updateEndpoint();
    }

    public static void setContext(String context)
    {
        serverContext = context;
        updateEndpoint();
    }

    public static void setEndpoint(String host, int port, String prefix)
    {
        serverPort = port;
        serverHost = host;
        serverContext = prefix;
        updateEndpoint();
    }

    public static String getBaseURI()
    {
        return baseURI;
    }

    public static String doMethod(int[] expect, HttpMethod method, String ... nvParams) throws HttpResponseException
    {
//        log.trace("Method: " + method.getName() + " " + method.getURI().toString() + " expect response: " + expect);
        int status = -1;
        int ex = expect.length == 0 ? -1 : expect[0];

        if (nvParams.length != 0)
        {
            NameValuePair[] params = new NameValuePair[nvParams.length];

            for (int i = 0; i < nvParams.length; i++)
            {
                String[] pair = nvParams[i].split("=");
                if (expect.length == 0)
                    assert pair.length == 2;
                params[i] = new NameValuePair(pair[0], pair[1]);
            }

            method.setQueryString(params);
        }

        try
        {
            log.trace(method.getName() + ": " + method.getURI());
            status = client.executeMethod(method);

            for (int i : expect)
                if (i == status) {
                    return getResponseBodyFromSteam(method);
                }

            throw new HttpResponseException(null, getResponseBodyFromSteam(method), ex, status);

        }
        catch (IOException e)
        {
            throw new HttpResponseException(e, "", ex, status);
        }
        finally
        {
            method.releaseConnection();
        }
    }

    private static String getResponseBody(HttpMethod method)
    {
        try
        {
            if (method.getResponseBody() == null)
                return method.getStatusLine().toString();

            return new String(method.getResponseBody(), "US-ASCII");
        }
        catch (IOException e)
        {
            return e.getMessage();
        }
    }

    private static String getResponseBodyFromSteam(HttpMethod method)
    {
        try
        {
            InputStream is = method.getResponseBodyAsStream();
            StringBuilder sb = new StringBuilder();
            byte[] buf = new byte[1024];
            int len;

            while ((len = is.read(buf)) != -1)
                sb.append(new String(buf, 0, len)); //, "US-ASCII");

            is.close();

            return sb.toString();
        }
        catch (IOException e)
        {
            return e.getMessage();
        }
    }

    public static String doGet(int[] expect, String url, String... nvParams) throws HttpResponseException
    {
        return doMethod(expect, new GetMethod(url), nvParams);
    }

    public static String doPost(int[] expect, String url, String... nvParams) throws HttpResponseException
    {
        HttpMethod method = new PostMethod(url);
        String res = doMethod(expect, method, nvParams);
        Header header = method.getResponseHeader("location");
        log.debug("post response location header: " + (header != null ? header.getValue() : "null"));
        return header != null ? header.getValue() : res;
    }

    public static String doPut(int[] expect, String url, String request, String... nvParams) throws HttpResponseException
    {
        PutMethod pm = new PutMethod(url);

        if (request != null) {
            try
            {
                pm.setRequestEntity(new StringRequestEntity(request, "text/html", null));
            }
            catch (UnsupportedEncodingException e)
            {
                log.debug("put request entity create error");
                e.printStackTrace();
            }
        }

        return doMethod(expect, pm, nvParams);
    }

    public static String doDelete(int[] expect, String url, String... nvParams) throws HttpResponseException
    {
        return doMethod(expect, new DeleteMethod(url), nvParams);
    }

    public static String stringFormat(Object obj) throws JAXBException
    {
        JettisonMappedContext context = new JettisonMappedContext(obj.getClass());
        StringWriter writer = new StringWriter();
        Marshaller marshaller = context.createMarshaller();
        marshaller.marshal(obj, writer);
        
        return writer.toString();
    }

    public static Response postResponse(String uri)
    {
        try
        {
            return Response.status(HttpResponseCodes.SC_CREATED).location(new URI(uri)).build();
        }
        catch (URISyntaxException e)
        {
            return Response.status(HttpResponseCodes.SC_INTERNAL_SERVER_ERROR).build();
        }
    }

    public static Response response(int code)
    {
        return Response.status(code).build();
    }

    public static Response response(String body)
    {
        if (body == null || body.length() == 0)
            return TxUtil.response(HttpResponseCodes.SC_NO_CONTENT);

        Response.ResponseBuilder builder = Response.ok(body);

        return builder.build();
    }
}
