package org.jboss.jbossts.rts.test;

import junit.framework.TestCase;

import org.jboss.resteasy.plugins.server.tjws.TJWSEmbeddedJaxrsServer;
import org.jboss.resteasy.util.HttpResponseCodes;
import org.jboss.jbossts.rts.util.TxUtil;
import org.jboss.jbossts.rts.service.TMApplication;
import org.jboss.jbossts.rts.provider.HttpResponseException;
import org.apache.log4j.Logger;

import java.io.IOException;

/**
 * Run through some basic tests to validate that the protocol matches the specification
 */
public class AppTest extends TestCase
{
    protected final static Logger log = Logger.getLogger(AppTest.class);
    private static int PORT = TxUtil.getIntProperty("PORT", "9096");
    private static TJWSEmbeddedJaxrsServer tjws;

    private String[] URLS;

    /**
     * Create the test case
     *
     * @param testName name of the test case
     */
    public AppTest( String testName )
    {
        super( testName );
    }

    protected void setUp() throws Exception
    {
        super.setUp();

        TxUtil.setEndpoint("localhost", PORT, "");

        log.trace("setUp: " + TxUtil.getBaseURI());
        URLS = new String[] {
                TxUtil.txURI + ".html", // list txns
                TxUtil.txURI + "/begin?clientId=12345",   // start a tx on behalf of client id 12345
                TxUtil.txURI + "/tx/commit",   // commit tx 123
                TxUtil.txURI + "/0/commit", // commit a non existant tx
                TxUtil.txURI + "/commit", // commit a non existant tx
                TxUtil.txURI + "/tx",   // enlist a participant in tx 123
        };

        tjws = new TJWSEmbeddedJaxrsServer();
        tjws.setPort(PORT);
        tjws.setBindAddress("127.0.0.1");
        TMApplication.initApplication(tjws);
        tjws.start();
        System.out.println("web server is running on port " + PORT);
    }

    protected void tearDown() throws Exception
    {
        log.trace("stoppig server");
        super.tearDown();

        tjws.stop();
    }

    private String startTx(long timeout, boolean fail) throws IOException
    {
        String tx = null;

        try {
            String query = "clientId=12345" + (timeout == 0 ? "" : "&timeout=" + timeout);

            tx = TxUtil.doPost(new int[] {HttpResponseCodes.SC_CREATED}, TxUtil.txURI + "/begin?" + query);
            log.trace("Started tx: " + tx);
            if (fail)
                assertTrue("start tx should have failed", false);
        } catch (HttpResponseException e) {
            if (!fail)
                assertTrue(e.getMessage(), false);
        }

        return tx;
    }

    private String modifyResource(String tx, String pid, String name, String value)
    {
        try {
            return TxUtil.doPost(new int[] {HttpResponseCodes.SC_OK}, TxUtil.getBaseURI() + "participant/work/" + pid,
                    "name=" + name, "value=" + value, "context=" + tx);
        } catch (HttpResponseException e) {
            assertTrue(e.getMessage(), false);
            return null;
        }
    }

    private String getResource(String tx, String pid, String name)
    {
        try {
            return TxUtil.doGet(new int[] {HttpResponseCodes.SC_OK}, TxUtil.getBaseURI() + "participant/work/" + pid,
                    "name=" + name, "context=" + tx);
        } catch (HttpResponseException e) {
            assertTrue(e.getMessage(), false);
            return null;
        }
    }

    private void terminateTx(String tx, String how, boolean fail, String ... args)
    {
        try {
            TxUtil.doPut(new int[] {HttpResponseCodes.SC_OK}, tx + "/" + how, "", args);
            if (fail)
                assertTrue("commit should not have succeeded", false);
        } catch (HttpResponseException e) {
            if (!fail)
                assertTrue(e.getMessage(), false);
	}
    }

    private void commitTx(String tx, boolean fail, String ... args)
    {
        terminateTx(tx, "commit", fail, args);
    }

    private void rollbackTx(String tx, boolean fail, String ... args)
    {
        terminateTx(tx, "rollback", fail, args);
    }

    // 1PC commit
    public void test1() throws IOException
    {
        String tx = startTx(0L, false);
        modifyResource(tx, "pid1", "p1", "v1");
        commitTx(tx, false);
    }
    // 2PC commit
    public void test2() throws IOException
    {
        String tx = startTx(0L, false);
        modifyResource(tx, "pid1", "p1", "v1");
        modifyResource(tx, "pid2", "p2", "v2");
        commitTx(tx, false);
        commitTx(tx, true);	// should not be able to commit a completed transaction
    }
    // commit an invalid transaction
    public void test3() throws IOException
    {
        String tx = "http://localhost:9096/transaction-coordinator/dead_7f000001_b9cc_4981ded5_5";
        commitTx(tx, true);
    }

    /**
     * Modify a resource within a transaction and rollback.
     * Ensure that the resource has not changed.
     */
    public void test4() throws IOException
    {
        String v1 = modifyResource(null, "pid1", "p1", "v1");
        String v2 = modifyResource(null, "pid2", "p1", "v1");

        String tx = startTx(0L, false);

        String v3 = modifyResource(tx, "pid1", "p1", "txv1");
        String v4 = modifyResource(tx, "pid2", "p1", "txv1");

        rollbackTx(tx, false);
        rollbackTx(tx, true);	// should not be able to rollback a completed transaction

        // make sure the changes were rolled back
        String v5 = getResource(null, "pid1", "p1");
        String v6 = getResource(null, "pid2", "p1");

        assertTrue("participant 1's work wasn't rolled back correctly", v1.equals(v5));
        assertTrue("participant 2's work wasn't rolled back correctly", v2.equals(v6));
    }

    /**
     * Test the Recovery URL
     */
    public void test5() throws IOException
    {
        String tx = startTx(0L, false);	// start a transaction

        String pUrl = TxUtil.getBaseURI() + "particpant/123";	// participant url
	// Registering a participant with the transaction returns a recovery url
        String recCoordId = TxUtil.doPut(new int[] {HttpResponseCodes.SC_OK}, tx, pUrl, "URL=" + pUrl);

	// doing a GET on recCoordId should return the original url
        String origUrl = TxUtil.doGet(new int[] {HttpResponseCodes.SC_OK}, recCoordId);
	assertEquals("GET on recovery url does not return orignal participant url", pUrl, origUrl);

	// doing a PUT on recCoordId should replace the participant URL (used by
	// participants to tell the recovery coordinator that they have moved
	pUrl = TxUtil.getBaseURI() + "particpant/123"; // the new participant URL
        // replace the url with a new one
        TxUtil.doPut(new int[] {HttpResponseCodes.SC_OK}, recCoordId, pUrl, "URL=" + pUrl);

        // check that the url was replaced by doing another GET on the recovery URL
        origUrl = TxUtil.doGet(new int[] {HttpResponseCodes.SC_OK}, recCoordId);
        assertEquals("GET on recovery url does not return the replaced participant url", pUrl, origUrl);

        // and finally check that POST and DELETE on the recovery URL fail
        TxUtil.doPost(new int[] {HttpResponseCodes.SC_UNAUTHORIZED}, recCoordId);
        TxUtil.doDelete(new int[] {HttpResponseCodes.SC_UNAUTHORIZED}, recCoordId);

        rollbackTx(tx, false);
    }

    /*
     * Test recovery
     */
    public void test6() throws IOException
    {
        boolean crashVM = false;
        String[] txns = {startTx(0L, false), startTx(0L, false)};

        for (int i = 0; i < txns.length; i++) {
            modifyResource(txns[i], "pid1", "p1", "v1" + i);
            modifyResource(txns[i], "pid2", "p2", "v2" + i);
        }

        // list transactions
        String txlist = TxUtil.doGet(new int[] {HttpResponseCodes.SC_OK}, TxUtil.txURI);
        // there should be 2
        assertEquals("List transactions does not match the number of running transactions",
            txlist.split("ActionStatus").length - 1, 2);

        for (int i = 0; i < txns.length; i++) {
            if (i == txns.length - 1 && crashVM)
            	commitTx(txns[i], false, "fault=commit_halt");
            else
            	commitTx(txns[i], false);
        }
    }
}
